/* Copyright (c) 2007-2010, Michal Bidlo (bidlom AT fit.vutbr.cz)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the organization nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <cmath>
#include <gtk/gtk.h>

#include "EvoSort.h"
#include "interface.h"

using namespace std;

////////////////////////////////////////////////////////////////////////////////
// program parameters

// comment DEBUG to evolve without showing the progress
#define DEBUG

// maximal sorting network width; larger values are recommended 
// only for testing of resulting evolved sorting networks
extern const UINT maxw = 12;
const UINT stepw = 4;   // development step width
// embryo: here basic 4-input sorting network is used
UINT embryo[][2] = { {0, 1}, {2, 3}, {0, 2}, {1, 3}, {1, 2} };
// according to the parameters, the fitness will evaluate 
// 8-input and 12-input sorting network developed from the embryo
////////////////////////////////////////////////////////////////////////////////


// rozmery pole pro vyvoj obvodu
extern const UINT cmplim = 1000;

// maximalni ciselna konstanta pripustna pro instrukce
const UINT max_const = 4;

UINT env[] = { 0 };

const UINT aprog_len[] = { 8 };
const UINT progs_len = 8;

const UINT vars = 6;
const UINT pmut = 4;
const UINT pcross = 100;
const UINT unit = 100;
const UINT tour = 2;
const UINT popsize = 32;

const UINT generations = 20000;

extern UINT sn_length;

bool prune_r;

enum ELEMENT
{
    v0 = max_const + 1, v1, v2, v3, v4, v5,       // promenne
    max_v
};

UINT var[vars];

// symboly instrukci
enum INST
{
    SET, INC, DEC, REP, GEN, NOP,
    INSS
};

struct Sblock
{
    UINT wide;
    UINT posvar;
};

Sblock block[] = {    // sirky komparatoru pripustne pro navrh
    { 1, v0 },
    { 1, v1 },
    { 1, v2 },
    { 1, v3 },
    { 2, v0 },
    { 2, v1 },
    { 2, v2 },
    { 2, v3 },
    { 3, v0 },
    { 3, v1 },
    { 3, v2 },
    { 3, v3 },
    { 4, v0 },
    { 4, v1 },
    { 4, v2 },
    { 4, v3 },
};

const UINT blocks = sizeof(block) / sizeof(block[0]);

Cell circuit[cmplim];   // pole pro vyvoj radici site
UINT pos;   // aktualni pozice generatoru
UINT embw;  // sirka embrya
UINT width; // aktualni sirka radici site

// ----------------------------- implementace GA -------------------------------

UINT offset[sizeof(aprog_len) / sizeof(aprog_len[0])];

struct GA_chromosome	// datovy typ chromozomu
{
    Inst prog[progs_len];
	double fitness;
	bool evaluate;
};

double max_fit;
double best_ever;
GA_chromosome best;	// nejlepsi dosud nalezene reseni
    
UINT generation;
GA_chromosome *population;
GA_chromosome *next_population;
GA_chromosome pool1[popsize];
GA_chromosome pool2[popsize];

// struktury potrebne pro evaluaci obvodu
const UINT patterns[]={0xAAAAAAAA,0xCCCCCCCC,0xF0F0F0F0,0xFF00FF00,0xFFFF0000};

UINT sequences[maxw + 1];
UINT vectors[maxw + 1];

ostream & operator<<(ostream &os, const Cell &c)
{
    return os;
}

ostream & operator<<(ostream &os, const Cell circ[cmplim])
{
    for (UINT x = 0; x < cmplim; x++)
      cout << "[" << circ[x].in1 << "," << circ[x].in2 << "," << circ[x].active << "]" << " "; 
    cout << endl;
    return os;
}

ostream & operator<<(ostream &os, const Inst &i)
{
#ifdef DEBUG
    switch (i.opc)
    {
        case SET: os << "SET "; break;
        case INC: os << "INC "; break;
        case DEC: os << "DEC "; break;
        case REP: os << "REP "; break;
        case GEN: os << "GEN "; break;
        case NOP: os << "NOP "; break;
        default: os << "Error in opcode!!" << endl; exit(i.opc); break;
    }
    if (i.opc == NOP)
        os << endl;
    else if (i.opc == GEN)
    {
        os << "C" << (UINT) block[i.arg1].wide;
        switch (block[i.arg1].posvar)
        {
            case v3: os << "v3" << endl; break;
            case v2: os << "v2" << endl; break;
            case v1: os << "v1" << endl; break;
            case v0: os << "v0" << endl; break;
            default: break;
        }     
    }
    else
    {
        switch (i.arg1)
        {
            case v5: os << "v5 "; break;
            case v4: os << "v4 "; break;
            case v3: os << "v3 "; break;
            case v2: os << "v2 "; break;
            case v1: os << "v1 "; break;
            case v0: os << "v0 "; break;
            default:
                if (i.arg1 < v0) os << i.arg1 << " ";
//                else os << "Error in arg1!!" << endl; exit(i.arg1); 
                break;
        }
        switch (i.arg2)
        {
            case v5: os << "v5" << endl; break;
            case v4: os << "v4" << endl; break;
            case v3: os << "v3" << endl; break;
            case v2: os << "v2" << endl; break;
            case v1: os << "v1" << endl; break;
            case v0: os << "v0" << endl; break;
            default:
                if (i.arg2 < v0) os << i.arg2 << endl;
//                else os << "Error in arg2!!" << endl; exit(i.arg2); 
                break;
        }
    }
#else
    os << i.opc << " " << i.arg1 << " " << i.arg2 << " ";
#endif

    return os;
}

ostream & operator<<(ostream &os, const GA_chromosome &g)
{
	for (UINT i = 0; i < progs_len; i++) os << g.prog[i];
    
	return os;
}

UINT random(UINT low, UINT high)
{
    return (rand() % (high - low + 1)) + low;
}

void set_args(Inst &i)
{
        switch (i.opc)
        {
            case SET:
                i.arg1 = random(v0, max_v - 1);
                i.arg2 = random(0, max_v - 1);
                break;
            case INC:
            case DEC:
                i.arg1 = random(v0, max_v - 1);
                i.arg2 = random(0, max_const);
                break;
            case REP:
                i.arg1 = random(v0, max_v - 1);
                i.arg2 = random(0, max_const);
                break;
            case GEN:
                i.arg1 = random(0, blocks - 1);
                i.arg2 = blocks; // random(0, blocks);
                break;
            case NOP:
                i.arg1 = i.arg2 = 0;
                break;
            default: break;
        }
}

void initialize(GA_chromosome &genome)
{
    for (UINT i = 0; i < progs_len; i++)
    {
        genome.prog[i].opc = random(0, INSS - 1);
        set_args(genome.prog[i]);
    }
}

void set(UINT store, UINT source)
{
    if (source <= max_const)
        var[store - v0] = source;    // prirazeni ciselne konstanty
    else
        var[store - v0] = var[source - v0]; // prirazeni hodnoty promenne
}

void inc(UINT _var, UINT _val)
{
    var[_var - v0] += _val;
}

void dec(UINT _var, UINT _val)
{
    if (var[_var - v0] >= _val)
        var[_var - v0] -= _val;
}

void gen(UINT _block1, UINT _block2)
{
    if (pos < cmplim &&
        var[block[_block1].posvar - v0] + block[_block1].wide < width)
    {
        circuit[pos].in1=var[block[_block1].posvar-v0];
        circuit[pos].in2=var[block[_block1].posvar-v0]+block[_block1].wide;
        circuit[pos].active = true;
        pos++;
    }
    if (_block2 < blocks && pos < cmplim &&
        var[block[_block2].posvar - v0] + block[_block2].wide < width)
    {
        circuit[pos].in1=var[block[_block2].posvar-v0];
        circuit[pos].in2=var[block[_block2].posvar-v0]+block[_block2].wide;
        circuit[pos].active = true;
        pos++;
    }
}

void rep(Inst *prog, UINT ip, UINT limit)
{
    UINT cnt;
    if (prog[ip].arg1 < v0)
        cnt = prog[ip].arg1;        // pocet opakovani dan konstantou
    else
        cnt = var[prog[ip].arg1 - v0];   // pocet opak. dan hodnotou promenne
    
    for (UINT i = 0; i < cnt; i++)	// arg1-krat opakovat arg2 instrukci za REP
    {
        for (UINT j = 1; ip + j < limit && j <= prog[ip].arg2; j++)
        {
            switch (prog[ip + j].opc)
            {
                case SET: set(prog[ip + j].arg1, prog[ip + j].arg2); break;
                case INC: inc(prog[ip + j].arg1, prog[ip + j].arg2); break;
                case DEC: dec(prog[ip + j].arg1, prog[ip + j].arg2); break;
                case GEN: gen(prog[ip + j].arg1, prog[ip + j].arg2); break;
                default: break;
            }
        }
    }
}

void run(Inst *prog, UINT limit)
{
    UINT ip = 0;
    while (ip < limit)
    {
        switch (prog[ip].opc)
        {
            case SET: set(prog[ip].arg1, prog[ip].arg2); ip++; break;
            case INC: inc(prog[ip].arg1, prog[ip].arg2); ip++; break;
            case DEC: dec(prog[ip].arg1, prog[ip].arg2); ip++; break;
            case REP: rep(prog, ip, limit); ip += prog[ip].arg2 + 1; break;
            case GEN: gen(prog[ip].arg1, prog[ip].arg2); ip++; break;
            default: ip++; break;
        }
    }
}

void develop(GA_chromosome &genome)
{
    var[0] = 0;
    var[1] = width - 2;
    var[2] = width - 2;
    var[3] = width - 3;
    var[4] = width - 4;
    var[5] = width;

    for (UINT i = 0; i < sizeof(env) / sizeof(env[0]) && pos < cmplim; i++)
        run(&genome.prog[offset[env[i]]], aprog_len[env[i]]);
}

double evaluate(UINT *input, UINT *result)
{
    UINT ss = ~0;
    UINT sho = 1;
    double corr = 0.0;
    
    for (UINT i = 0; i < maxw - 1; i++)
	ss &= ~result[i] | result[i + 1];
    while (ss)
    {
	if (ss & 1)	// test priznaku serazeni
	{
	    UINT ones = 0;
	    for (UINT i = 0; i < maxw; i++)
		if (input[i] & sho) ones++;
	    // vypocet poctu spravnych bitu vysledku
	    for (UINT i = 0; i < maxw; i++)
	    {
		if (i < maxw - ones)
		{
		    if (!(result[i] & sho))
			corr++;
		}
		else
		{
		    if (result[i] & sho)
			corr++;
		}
	    }
	}
	sho <<= 1;
	ss >>= 1;
    }
    
    return corr;
}

int depth(Cell sort_net[])      
{
        int layers = 0, procc = 0;
        bool *skip, *wires = new bool[maxw];

        skip = new bool[cmplim];
        for (UINT i = 0; i < cmplim; i++) skip[i] = false;

        while (procc < cmplim)
        {
                        for (UINT i = 0; i < maxw; i++)
                          wires[i] = false;
                        for (UINT i = 0; i < cmplim; i++)
                        {
                          if (!skip[i])
  			  {
                            if (!sort_net[i].active)
  			    {
			      skip[i] = true;
			      procc++;
			    }
			    else
			    {
			      if (!wires[sort_net[i].in1] && !wires[sort_net[i].in2])
                              {
                                skip[i] = true;
                                procc++;
                              }
                              wires[sort_net[i].in1] = true;
                              wires[sort_net[i].in2] = true;
			    }  
			  }     
                        }
                        layers++;
        }

        delete[] skip;
        delete[] wires;

        return layers;
}

double fitness(GA_chromosome &genome, UINT *cmps)
{
    double fit = 0.0;
    	
    // inicializace struktury site - embryo a vsechny dalsi bloky neaktivni
    pos = 0; embw = 0;
    sn_length = 0;
    bool rcomps[cmplim];
    
    for (UINT i = 0; i < cmplim; i++)
    {
        if (i < sizeof(embryo) / sizeof(embryo[0]))
        {
            if (embryo[i][1] > embw) embw = embryo[i][1];
            circuit[i].in1 = embryo[i][0];
            circuit[i].in2 = embryo[i][1];
            circuit[i].active = true;
            pos++;
        } else circuit[i].active = false;
        rcomps[i] = true;
    }
    embw = embw + 1;    // doted tam byl jen index posledniho dratu
    width = embw + stepw;
    
    for (width = embw + stepw; width <= maxw; width += stepw)
        develop(genome);

    UINT test[maxw], inputs[maxw];
    UINT a = vectors[maxw] - 1, p;
    for (UINT i = 0; i < vectors[maxw]; i++)
    {
		p = a;
		// vytvoreni testovacich sekvenci v podobe 1 unsigned integeru
		for (UINT j = 0; j < maxw; j++)
		{
		    if (j < 5) test[j] = patterns[j];
	    	else { test[j] = p & 1 ? 0xFFFFFFFF : 0; p >>= 1; }
		    inputs[j] = test[j];
		}
		a--;
        
		// simulace vypoctu obvodu s testovacimi sekvencemi na jeho vstupech
        for (UINT p = 0; p < cmplim; p++)
        {
            if (circuit[p].active)
	    {
              UINT t1 = test[circuit[p].in1] & test[circuit[p].in2];
              UINT t2 = test[circuit[p].in1] | test[circuit[p].in2];
              if ((t1 != test[circuit[p].in1]))  rcomps[p] = false;
              test[circuit[p].in1] = t1;
              test[circuit[p].in2] = t2;
	    }  
        }
		fit += evaluate(inputs, test);
    }
	
    if (prune_r)
      for (UINT p = 0; p < cmplim; p++)
        if (rcomps[p])
          circuit[p].active = false;
	    
    // Count comparators if needed
    if (cmps != NULL)
    {
       *cmps = 0;
       for (UINT snl = 0; snl < cmplim; snl++)
         if (circuit[snl].active)
	   (*cmps)++;
       sn_length = *cmps - 1;
    }	   
    return fit;
}

void crossover(GA_chromosome &parent1, GA_chromosome &parent2,
		GA_chromosome &child1, GA_chromosome &child2)
{
    child1 = parent1;
    child2 = parent2;
    
    child1.prog[random(0, progs_len-1)] = parent2.prog[random(0, progs_len-1)];
    child2.prog[random(0, progs_len-1)] = parent1.prog[random(0, progs_len-1)];
}

bool mutator(GA_chromosome &genome, UINT _pmut)
{
    bool mutated = false;

    if (random(0, unit) < _pmut)
    {
        mutated = true;
        UINT in = random(0, progs_len - 1); // kterou instrukci mutovat
        UINT part = random(0, unit);    // jakou cast instrukce mutovat
        
        if (part <= 33)   // muatace cele instrukce
        {
            genome.prog[in].opc = random(0, INSS - 1);
            set_args(genome.prog[in]);
        }
        else if (part <= 66)    // mutace arg1
        {
            // mutaci argumentu s ohledem na typ instrukce
            switch (genome.prog[in].opc)
            {
                case SET: case INC: case DEC:
                    genome.prog[in].arg1 = random(v0, max_v - 1); break;
                case REP: genome.prog[in].arg1 = random(0, max_v - 1); break;
                case GEN: genome.prog[in].arg1 = random(0, blocks - 1); break;
                default: break;
            }
        }
        else    // mutace arg2
        {
            switch (genome.prog[in].opc)
            {
                case SET: genome.prog[in].arg2 = random(0, max_v - 1); break;
                case INC: case DEC: case REP:
                    genome.prog[in].arg2 = random(0, max_const); break;
                case GEN: genome.prog[in].arg2 = blocks; break; //random(0, blocks); break;
                default: break;
            }
        }
    }

    return mutated;
}

bool stop()
{
    if (best.fitness > best_ever)
    {
#ifdef DEBUG
		cout << "Fitness = " << best.fitness << "/" << max_fit
	        << " in generation " << generation << endl;
#endif
		best_ever = best.fitness;
    }
	
    if (best_ever == max_fit || generation == generations)
	{
		ofstream ofs("results", ios_base::app);
		if (best_ever == max_fit)
        	    ofs << "gen. " << generation << endl << best << endl;
		else ofs << "FAILED" << endl;
		ofs.close();
		return true;
	}
    
    return false;
}

// evolucni cyklus: nova populace o popsize jedincich je tvorena krizenimi
// a mutaci turnajem vybranymi jedinci predchazejici generace
// specifikum: elitismus nahrazuje jedince s totoznymi fitness - vyrazny uspech!
void evolve()
{
	// inicializace
	generation = 0;
	best.fitness = 0;
	best_ever = 0.0;
	GA_chromosome ind1_new, ind2_new;
	UINT _tour = tour >= 2 ? tour : 2;
	UINT i1;
    
	for (UINT i = 0; i < popsize; i++)
	{
		initialize(pool1[i]);
		pool1[i].evaluate = true;
	}
	// evolucni cyklus
	do
	{
		generation++;
		if (generation & 1)
		{
			population = pool1;
			next_population = pool2;
		}
		else
		{
			population = pool2;
			next_population = pool1;
		}
		// ohodnoceni populace
		for (UINT i = 0; i < popsize; i++)
		{
			if (population[i].evaluate)
			{
				population[i].fitness = fitness(population[i], NULL);
				if (population[i].fitness >= best.fitness) best = population[i];
				population[i].evaluate = false;
			}
		}
		// elitizmus
		next_population[0] = best;
        GA_chromosome mutant = best; mutator(mutant, unit);
		next_population[1] = mutant;
		// tvorba nove populace
		for (UINT i = 2; i < popsize; i += 2)
		{
			GA_chromosome *ind1 = NULL, *ind2 = NULL;
			// turnajovy vyber jedincu
			for (UINT t = 0; t < _tour; t++)
			{
				i1 = random(0, popsize - 1);
				if (ind1 == NULL) ind1 = &population[i1];
				else if (ind2 == NULL) ind2 = &population[i1];
				else if (population[i1].fitness > ind1->fitness)
					ind1 = &population[i1];
				else if (population[i1].fitness > ind2->fitness)
					ind2 = &population[i1];
			}
			// krizeni
			if (random(0, unit) < pcross)
			{
				crossover(*ind1, *ind2, ind1_new, ind2_new);
				ind1_new.evaluate = true;
				ind2_new.evaluate = true;
			}
			else	// prechod jedincu bez krizeni
			{
				ind1_new = *ind1;
				ind2_new = *ind2;
			}
			// mutace
			if (mutator(ind1_new, pmut)) ind1_new.evaluate = true;
			if (mutator(ind2_new, pmut)) ind2_new.evaluate = true;
			// umisteni potomku do nove populace
			next_population[i] = ind1_new;
			next_population[i + 1] = ind2_new;
		}
	} while (!stop());
}

void system_init()
{
    // inicializace offsetu programu genomu
    offset[0] = 0;
    for (UINT i = 1; i < sizeof(aprog_len) / sizeof(aprog_len[0]); i++)
        offset[i] = offset[i - 1] + aprog_len[i - 1];
    
    // inicializace poctu testovacich vektoru
    for (UINT i = 2; i <= maxw; i++)
    {
		if (i < 6) sequences[i] = 32;
		else sequences[i] = (UINT) pow(2.0, (int) i);
		vectors[i] = (UINT) sequences[i] / (8 * sizeof(UINT));
    }
	max_fit = (double) sequences[maxw] * (double) maxw;
}

void help(char name[])
{
	   cout << "usage: " << name << " [FILENAME]" << endl;
	   cout << "If FILENAME is supplied, the system executes the evolved programs" << endl;
	   cout << "contained in the file, otherwise evolution is started." << endl;
}

int main(int argc, char *argv[])
{
    system_init();

    if (argc > 1)   // test vyevolvovanych konstruktoru
    {
        char n;
        GtkWidget *window;
                
        gtk_set_locale ();                              
        gtk_init (&argc, &argv);
        
        window = create_window ();
        gtk_widget_show (window);
        if (strncmp(argv[1], "--help", 65536) == 0)
	{
           help(argv[0]);
	   return 0;
	}  
	
	if (argc > 3)
	{
	   cerr << "Error: Invalid number of arguments" << endl;
	   help(argv[0]);
	   return -1;
	}
	
        n = 1;
        prune_r = false;
	
	if (argc == 3)
        {
	  if (strncmp(argv[1], "-p", 65536) == 0)
	  {
	    prune_r = true;
	    n = 2;
	  }
	  else
	  {
	    cerr << "Error: Invalid argument(s)" << endl;
	    help(argv[0]);
	    return -2;
	  }
	}  
	    
	ifstream ifs(argv[n]);
	
	if (!ifs)
	{
          cerr << "Unable to find file: " << argv[n] << endl;
	  return -3;
        }
        
	GA_chromosome g;
        UINT order = 1;
        UINT cmps;
        ifs >> g.prog[0].opc >> g.prog[0].arg1 >> g.prog[0].arg2;
        while (!ifs.eof())
        {
            for (UINT i = 1; i < progs_len; i++)
                ifs >> g.prog[i].opc >> g.prog[i].arg1 >> g.prog[i].arg2;
            if (fitness(g, &cmps) == max_fit)
            {
                cout << g << endl;
                cout << circuit << endl;
                cout << "comparators = " << cmps << endl;
                cout << "depth = " << depth(circuit) << endl;
                cout << "press space..." << endl;
		        gtk_main();
            }
            
            ifs >> g.prog[0].opc >> g.prog[0].arg1 >> g.prog[0].arg2;
            order++;
        }
        cout<<"No. of nets: "<<order-1<<endl;
        ifs.close();
    }
    else
    {
        while (1)
        {
            FILE *rand = fopen("/dev/random", "r");
            UINT r;
            fread(&r, sizeof(r), 1, rand);
            fclose(rand);
            srand(r);
            evolve();
        }
    }
    
	return 0;
}
