/* Copyright (c) 2007-2010, Michal Bidlo (bidlom AT fit.vutbr.cz)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the organization nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <iostream>
#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "EvoSort.h"

extern UINT maxw;
UINT sn_length;
extern UINT cmplim;
extern Cell circuit[];

const int sx = 20;
const int sy = 20;
const int l_space = 15;
const int c_space = 10;
const int n_space = 10;
const int cmpx_d = 10;

int last_y;

gboolean
on_window_delete_event (GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
	gtk_main_quit();
	
	return TRUE;
}

void draw_with_layers(GdkGC *gc, GtkWidget *widget)
{
	int x = 0, l = 0, l_length = 0;
	bool new_layer = TRUE;
	bool area[sn_length][maxw], layers[sn_length][maxw];
	for (int i = 0; i < sn_length; i++)
		for (int j = 0; j < maxw; j++)
		{
			area[i][j] = FALSE;
			layers[i][j] = FALSE;
		}
		
	for (int i = 0; i < cmplim; i++)
	{
            if (circuit[i].active)
	    {      
		int ax = x;
		bool free = TRUE;
		while (free && x > 0)
		{
			for (int j = circuit[i].in1; free && j <= circuit[i].in2; j++)
				if (area[x - 1][j]) free = FALSE;
			if (free) x--;
		}
		
		for (int j = circuit[i].in1; j <= circuit[i].in2; j++)
			area[x][j] = TRUE;
		
		if (x > l_length) l_length = x;
		
		gdk_draw_line(widget->window, gc,
			sx + cmpx_d + x * c_space, sy + l_space * circuit[i].in1,
			sx + cmpx_d + x * c_space, sy + l_space * circuit[i].in2);
		gdk_draw_arc(widget->window, gc, TRUE,
			sx + cmpx_d + x * c_space - 4, sy + l_space * circuit[i].in1 - 4,
			7, 7, 0, 64 * 360);
		gdk_draw_arc(widget->window, gc, TRUE,
			sx + cmpx_d + x * c_space - 4, sy + l_space * circuit[i].in2 - 4,
			7, 7, 0, 64 * 360);
		
		x = ax + 1;
	   }	
	}

	for (int i = 0; i < maxw; i++)
		gdk_draw_line(widget->window, gc,
			sx, sy + i * l_space,
			sx + c_space * l_length+ 2 * cmpx_d, sy + i * l_space);

	gtk_window_resize(GTK_WINDOW(gtk_widget_get_parent(widget)),
		2 * sx + c_space * l_length + 2 * cmpx_d, 2 * sy + l_space * (maxw - 1));
}

gboolean
draw(GtkWidget *widget, GdkEventExpose *event, gpointer user_data)
{
	GdkGC *gc = gdk_gc_new(widget->window);
	GdkColor color = {~0, 65535, 65535, 65535};
	gtk_widget_modify_bg(widget, GTK_STATE_NORMAL, &color);
	
        gdk_gc_set_line_attributes(gc, 2, GDK_LINE_SOLID, GDK_CAP_BUTT,
                GDK_JOIN_MITER);
        draw_with_layers(gc, widget);
 
	return TRUE;
}
